/*

    #     ####            ####
    #    #    #          #    #
    #    #    #          #
    #    #    #   ###    #
    #    #    #   ###    #    #
    #     ####    ###     ####

	I/O routines.
*/

/*
 * $Id: io.c,v 2.9 92/07/14 16:48:13 ram Exp $
 *
 *  Copyright (c) 1992, Raphael Manfredi
 *
 *  You may redistribute only under the terms of the GNU General Public
 *  Licence as specified in the README file that comes with dist.
 *
 * $Log:	io.c,v $
 * Revision 2.9  92/07/14  16:48:13  ram
 * 3.0 beta baseline.
 * 
 */

#include "config.h"
#include "portable.h"
#include <sys/types.h>
#include "hash.h"
#include "parser.h"
#include "lock.h"
#include "logfile.h"
#include "environ.h"
#include <stdio.h>
#include <errno.h>
#include <sys/stat.h>

#ifdef I_SYSWAIT
#include <sys/wait.h>
#endif

#ifdef I_FCNTL
#include <fcntl.h>
#else
#include <sys/fcntl.h>
#endif
#ifdef I_SYSFILE
#include <sys/file.h>
#endif

#ifdef I_STRING
#include <string.h>
#else
#include <strings.h>
#endif

#define BUFSIZE		1024			/* Amount of bytes read in a single call */
#define CHUNK		(10 * BUFSIZE)	/* Granularity of pool */
#define MAX_STRING	2048			/* Maximum string's length */
#define AGENT_WAIT	"agent.wait"	/* File listing out-of-the-queue mails */
#define AGENT_LOCK	"perl.lock"		/* Lock file used by mailagent */

private void pool_realloc();	/* Extend pool zone */
private int get_lock();			/* Attempt to get a lockfile */
private void release_agent();	/* Remove mailagent's lock if needed */
private int process_mail();		/* Process mail by feeding the mailagent */
private void queue_mail();		/* Queue mail for delayed processing */
private char *write_file();		/* Write mail on disk */
private char *save_file();		/* Emergency saving into a file */

private char *mail = (char *) 0;	/* Where mail is stored */
private int len;					/* Mail length in bytes */
private int queued = 0;				/* True when mail queued safely */

extern int errno;				/* System call error status */
extern char *malloc();			/* Memory allocation */
extern char *realloc();			/* Re-allocation of memory pool */
extern char *logname();			/* User's login name */
extern int loglvl;				/* Logging level */

private void read_stdin()
{
	/* Read the whole stdandard input into memory and return a pointer to its
	 * location in memory. Any I/O error is fatal. Set the length of the
	 * data read into 'len'.
	 */

	int size;					/* Current size of memory pool */
	int amount = 0;				/* Total amount of data read */
	int n;						/* Bytes read by last system call */
	char *pool;					/* Where input is stored */
	char buf[BUFSIZE];

	size = CHUNK;
	pool = malloc(size);
	if (pool == (char *) 0)
		fatal("out of memory");

	add_log(19, "reading mail");

	while (n = read(0, buf, BUFSIZE)) {
		if (n == -1) {
			add_log(1, "SYSERR read: %m (%e)");
			fatal("I/O error");
		}
		if (size - amount < n)				/* Pool not big enough */
			pool_realloc(&pool, &size);		/* Resize it or fail */
		bcopy(buf, pool + amount, n);		/* Copy read bytes */
		amount += n;						/* Update amount of bytes read */
	}

	len = amount;				/* Indicate how many bytes where read */

	add_log(16, "got mail (%d bytes)", amount);

	mail = pool;				/* Where mail is stored */
}

public void process()
{
	char *queue;						/* Location of mailagent's queue */

	(void) umask(077);					/* Files we create are private ones */

	queue = ht_value(&symtab, "queue");	/* Fetch queue location */
	if (queue == (char *) 0)
		fatal("queue directory not defined");

	read_stdin();						/* Read mail */
	(void) get_lock();					/* Get a lock file */
	queue_mail(queue);					/* Process also it locked */
	release_lock();						/* Release lock file if necessary */
}

public int was_queued()
{
	return queued;			/* Was mail queued? */
}

private void pool_realloc(pool, size)
char **pool;
int *size;
{
	/* Make more room in pool and update parameters accordingly */

	char *cpool = *pool;	/* Current location */
	int csize = *size;		/* Current size */

	csize += CHUNK;
	cpool = realloc(cpool, csize);
	if (cpool == (char *) 0)
		fatal("out of memory");
	*pool = cpool;
	*size = csize;
}

private int get_lock()
{
	/* Try to get a filter lock in the spool directory. Propagate the return
	 * status of filter_lock(): 0 for ok, -1 for failure.
	 */

	char *spool;						/* Location of spool directory */

	spool = ht_value(&symtab, "spool");	/* Fetch spool location */
	if (spool == (char *) 0)
		fatal("spool directory not defined");

	return filter_lock(spool);			/* Get a lock in spool directory */
}

private void release_agent()
{
	/* In case of abnormal failure, the mailagent may leave its lock file
	 * in the spool directory. Remove it if necessary.
	 */

	char *spool;					/* Location of spool directory */
	char agentlock[MAX_STRING];		/* Where lock file is held */
	struct stat buf;				/* Stat buffer */

	spool = ht_value(&symtab, "spool");	/* Fetch spool location */
	if (spool == (char *) 0)			/* Should not happen */
		return;

	sprintf(agentlock, "%s/%s", spool, AGENT_LOCK);
	if (-1 == stat(agentlock, &buf))
		return;						/* Assume no lock file left behind */

	if (-1 == unlink(agentlock)) {
		add_log(1, "SYSERR unlink: %m (%e)");
		add_log(2, "ERROR could not remove mailagent's lock");
	} else
		add_log(5, "NOTICE removed mailagent's lock");
}

private void queue_mail(queue)
char *queue;				/* Location of the queue directory */
{
	char *where;			/* Where mail is stored */
	char real[MAX_STRING];	/* Real queue mail */
	char *base;				/* Pointer to base name */
	struct stat buf;		/* To make sure queued file remains */

	where = write_file(queue, "Tm");
	if (where == (char *) 0) {
		add_log(1, "ERROR unable to queue mail");
		fatal("try again later");
	}

	/* If we have a lock, create a qm* file suitable for mailagent processing.
	 * Otherwise, create a fm* file and the mailagent will process it
	 * immediately.
	 */
	if (is_locked())
		sprintf(real, "%s/%s%d", queue, "qm", progpid);
	else
		sprintf(real, "%s/%s%d", queue, "fm", progpid);

	if (-1 == rename(where, real)) {
		add_log(1, "SYSERR rename: %m (%e)");
		add_log(2, "ERROR could not rename %s into %s", where, real);
		fatal("try again later");
	}

	/* Compute base name of queued mail */
	base = rindex(real, '/');
	if (base++ == (char *) 0)
		base = real;

	add_log(4, "QUEUED [%s] %d bytes", base, len);
	queued = 1;

	/* If we got a lock, then no mailagent is running and we may process the
	 * mail. Otherwise, do nothing. The mail will be processed by the currently
	 * active mailagent.
	 */

	if (!is_locked())			/* Another mailagent is running */
		return;					/* Leave mail in queue */

	if (0 == process_mail(real)) {
		/* Mailagent may have simply queued the mail for itself by renaming
		 * it, so of course we would not be able to remove it. Hence the
		 * test for ENOENT to avoid error messages when the file does not
		 * exit any more.
		 */
		if (-1 == unlink(real) && errno != ENOENT) {
			add_log(1, "SYSERR unlink: %m (%e)");
			add_log(2, "ERROR could not remove queued mail");
		}
		return;
	}
	/* Paranoia: make sure the queued mail is still there */
	if (-1 == stat(real, &buf)) {
		queued = 0;			/* Or emergency_save() would not do anything */
		add_log(1, "SYSERR stat: %m (%e)");
		add_log(1, "ERROR queue file [%s] vanished", base);
		if (-1 == emergency_save())
			add_log(1, "ERROR mail probably lost");
	} else {
		add_log(4, "WARNING mailagent failed, [%s] left in queue", base);
		release_agent();	/* Remove mailagent's lock file if needed */
	}
}

private int process_mail(location)
char *location;
{
	/* Process mail held in 'location' by invoking the mailagent on it. If the
	 * command fails, return -1. Otherwise, return 0;
	 * Note that we will exit if the first fork is not possible, but that is
	 * harmless, because we know the mail was safely queued, otherwise we would
	 * not be here trying to make the mailagent process it.
	 */
	
	FILE *fp;				/* The file pointer on pipe */
	char cmd[MAX_STRING];	/* The built command */
	char buf[MAX_STRING];	/* To store output from mailagent */
	char **envp;			/* Environment pointer */
#ifdef UNION_WAIT
	union wait status;		/* Waiting status */
#else
	int status;				/* Status from command */
#endif
	int xstat;				/* The exit status value */
	int pid;				/* Pid of our children */
	int res;				/* Result from wait */

	if (loglvl <= 20) {		/* Loggging level higher than 20 is for tests */
		pid = fork();
		if (pid == -1) {	/* Resources busy, most probably */
			release_lock();
			add_log(1, "SYSERR fork: %m (%e)");
			add_log(6, "NOTICE exiting to save resources");
			exit(0);		/* Exiting will also release sendmail process */
		} else if (pid != 0)
			exit(0);		/* Release waiting sendmail */
	}

	/* Now hopefully we detached ourselves from sendmail, which thinks the mail
	 * has been delivered. Not yet, but close. Simply wait a little in case
	 * more mail is comming. This process is going to remain alive while the
	 * mailagent is running so as to trap any weird exit status. But the size
	 * of the perl process (with script compiled) is about 1650K on my MIPS,
	 * so the more we delay the invocation, the better.
	 */

	if (loglvl < 12)		/* Loggging level 12 and higher is for debugging */
		sleep(60);			/* Delay invocation of mailagent */
	progpid = getpid();		/* This may be the child (if fork succeded) */
	envp = make_env();		/* Build new environment */

	pid = vfork();			/* Virtual fork this time... */
	if (pid == -1) {
		add_log(1, "SYSERR vfork: %m (%e)");
		add_log(1, "ERROR cannot run mailagent");
		return -1;
	}

	if (pid == 0) {			/* This is the child */
		execle(PERLPATH, "perl", "-S", "mailagent", location, (char *) 0, envp);
		exit(1);
	} else {				/* Parent process */
		while (pid != (res = wait(&status)))
			if (res == -1) {
				add_log(1, "SYSERR wait: %m (%e)");
				return -1;
			}

#ifdef WEXITSTATUS
		if (WIFEXITED(status)) {			/* Exited normally */
			xstat = WEXITSTATUS(status);
			if (xstat != 0) {
				add_log(3, "ERROR mailagent returned status %d", xstat);
				return -1;
			}
		} else if (WIFSIGNALED(status)) {	/* Signal received */
			xstat = WTERMSIG(status);
			add_log(3, "ERROR mailagent terminated by signal %d", xstat);
			return -1;
		} else if (WIFSTOPPED(status)) {	/* Process stopped */
			xstat = WSTOPSIG(status);
			add_log(3, "WARNING mailagent stopped by signal %d", xstat);
			add_log(6, "NOTICE terminating mailagent, pid %d", pid);
			if (-1 == kill(pid, 15))
				add_log(1, "SYSERR kill: %m (%e)");
			return -1;
		} else
			add_log(1, "BUG please report bug 'posix-wait' to author");
#else
#ifdef UNION_WAIT
		xstat = status.w_status;
#else
		xstat = status;
#endif
		if ((xstat & 0xff) == 0177) {		/* Process stopped */
			xstat >>= 8;
			add_log(3, "WARNING mailagent stopped by signal %d", xstat);
			add_log(6, "NOTICE terminating mailagent, pid %d", pid);
			if (-1 == kill(pid, 15))
				add_log(1, "SYSERR kill: %m (%e)");
			return -1;
		} else if ((xstat & 0xff) != 0) {	/* Signal received */
			xstat &= 0xff;
			if (xstat & 0200) {				/* Dumped a core ? */
				xstat &= 0177;
				add_log(3, "ERROR mailagent dumped core on signal %d", xstat);
			} else
				add_log(3, "ERROR mailagent terminated by signal %d", xstat);
			return -1;
		} else {
			xstat >>= 8;
			if (xstat != 0) {
				add_log(3, "ERROR mailagent returned status %d", xstat);
				return -1;
			}
		}
#endif
	}
	
	add_log(19, "mailagent ok");

	return 0;
}

public int emergency_save()
{
	/* Save mail in emeregency files and add the path to the agent.wait file,
	 * so that the mailagent knows where to look when processing its queue.
	 * Return -1 if the mail was not sucessfully saved, 0 otherwise.
	 */

	char *where;			/* Where file was stored (static data) */
	char *home = homedir();	/* Location of the home directory */
	char path[MAX_STRING];	/* Location of the AGENT_WAIT file */
	char *queue;			/* Location of the queue directory */
	char *emergdir;			/* Emergency directory */
	int fd;					/* File descriptor to write in AGENT_WAIT */
	int size;				/* Length of 'where' string */

	if (mail == (char *) 0)
		return -1;			/* Mail not read yet */

	if (queued) {
		add_log(6, "NOTICE mail was safely queued");
		return 0;
	}

	emergdir = ht_value(&symtab, "emergdir");
	if ((emergdir != (char *) 0) && (char *) 0 != (where = save_file(emergdir)))
		goto ok;
	if ((home != (char *) 0) && (char *) 0 != (where = save_file(home)))
		goto ok;
	if (where = save_file("/usr/spool/uucppublic"))
		goto ok;
	if (where = save_file("/var/spool/uucppublic"))
		goto ok;
	if (where = save_file("/usr/tmp"))
		goto ok;
	if (where = save_file("/var/tmp"))
		goto ok;
	if (where = save_file("/tmp"))
		goto ok;

	return -1;		/* Could not save mail anywhere */

ok:
	add_log(6, "DUMPED in %s", where);
	fprintf(stderr, "%s: DUMPED in %s\n", progname, where);

	/* Attempt to write path of saved mail in the AGENT_WAIT file */

	queue = ht_value(&symtab, "queue");
	if (queue == (char *) 0)
		return 0;
	sprintf(path, "%s/%s", queue, AGENT_WAIT);
	if (-1 == (fd = open(path, O_WRONLY | O_APPEND | O_CREAT, 0600))) {
		add_log(1, "SYSERR open: %m (%e)");
		add_log(6, "WARNING mailagent ignores where mail was left");
		return 0;
	}
	size = strlen(where);
	where[size + 1] = '\0';			/* Make room for trailing new-line */
	where[size] = '\n';
	if (-1 == write(fd, where, size + 1)) {
		add_log(1, "SYSERR write: %m (%e)");
		add_log(4, "ERROR could not append to %s", path);
		add_log(6, "WARNING mailagent ignores where mail was left");
	} else {
		where[size] = '\0';
		add_log(7, "NOTICE memorized %s", where);
		queued = 1;
	}
	close(fd);

	return 0;
}

private char *save_file(dir)
char *dir;				/* Where saving should be done (directory) */
{
	/* Attempt to write mail in directory 'dir' and return a pointer to static
	 * data holding the path name of the saved file if writing was ok.
	 * Otherwise, return a null pointer and unlink any already created file.
	 */

	struct stat buf;				/* Stat buffer */

	/* Make sure 'dir' entry exists, although we do not make sure it is really
	 * a directory. If 'dir' is in fact a file, then open() will loudly
	 * complain. We only want to avoid spurious log messages.
	 */

	if (-1 == stat(dir, &buf))		/* No entry in file system, probably */
		return (char *) 0;			/* Saving failed */

	return write_file(dir, logname());
}

private char *write_file(dir, template)
char *dir;				/* Where saving should be done (directory) */
char *template;			/* First part of the file name */
{
	/* Attempt to write mail in directory 'dir' and return a pointer to static
	 * data holding the path name of the saved file if writing was ok.
	 * Otherwise, return a null pointer and unlink any already created file.
	 * The file 'dir/template.$$' is created (where '$$' refers to the pid of
	 * the current process). As login name <= 8 and pid is <= 5, we are below
	 * the fatidic 14 chars limit for filenames.
	 */

	static char path[MAX_STRING];	/* Path name of created file */
	int fd;							/* File descriptor */
	register4 int n;				/* Result from the write system call */
	register1 char *mailptr;		/* Pointer into mail buffer */
	register2 int length;			/* Number of bytes already written */
	register3 int amount;			/* Amount of bytes written by last call */

	sprintf(path, "%s/%s.%d", dir, template, progpid);

	if (-1 == (fd = open(path, O_WRONLY | O_CREAT | O_EXCL, 0600))) {
		add_log(1, "SYSERR open: %m (%e)");
		add_log(2, "ERROR cannot create file %s", path);
		return (char *) 0;
	}

	/* Write the mail on disc. We do not call a single write on the mail buffer
	 * as in "write(fd, mail, len)" in case the mail length exceeds the maximum
	 * amount of bytes the system can atomically write.
	 */
	
	for (
		mailptr = mail, length = 0;
		length < len;
		mailptr += amount, length += amount
	) {
		amount = len - length;
		if (amount > BUFSIZ)		/* Do not write more than BUFSIZ */
			amount = BUFSIZ;
		n = write(fd, mailptr, amount);
		if (n == -1 || n != amount) {
			if (n == -1)
				add_log(1, "SYSERR write: %m (%e)");
			add_log(2, "ERROR cannot write to file %s", path);
			if (-1 == unlink(path)) {
				add_log(1, "SYSERR unlink: %m (%e)");
				add_log(4, "WARNING leaving %s around", path);
			}
			close(fd);
			return (char *) 0;
		}
	}

	close(fd);
	add_log(19, "mail in %s", path);

	return path;			/* Where mail was writen (static data) */
}

#ifndef RENAME
public int rename(from, to)
char *from;				/* Original name */
char *to;				/* Target name */
{
	(void) unlink(to);
	if (-1 == link(from, to))
		return -1;
	if (-1 == unlink(from))
		return -1;

	return 0;
}
#endif

